from __future__ import division
from builtins import zip
from builtins import range
from pybasicbayes.abstractions import GibbsSampling
import numpy as np
from warnings import warn
from pypolyagamma import PyPolyaGamma
__all__ = ['Dynamic_GLM']


def local_multivariate_normal_draw(x, sigma, normal):
    """
    Function to combine pre-drawn Normals (normal) with the desired mean x and variance sigma (calls to multivariate normal are too costly)
    Cholesky doesn't like 0 cov matrix, but we want it.

    This might need changing if in practice we see plently of 0 matrices
    """
    try:
        return x + np.linalg.cholesky(sigma).dot(normal)
    except np.linalg.LinAlgError:
        if np.isclose(sigma, 0).all():
            return x
        else:
            print("Unusual covariance matrix")
            quit()


# fix seed of PyPolyaGamma, otherwise results aren't reproducible
ppgsseed = 4
if ppgsseed == 4:
    print("Using default seed")
ppgs = PyPolyaGamma(ppgsseed)


class Dynamic_GLM(GibbsSampling):
    """
    This class enables a drifting input output GLM with logistic link function.

    States are thus dynamic logistic regressions. This class implements this observation distribution for a single state.

    Hyperparameters:

        n_regressors: number of regressors for the GLM
        T: number of timesteps (sessions)
        prior_mean: mean of regressors at the beginning (usually 0 vector)
        P_0: variance of regressors at the beginning
        Q: variance of regressors between timesteps (can be different across steps, but we use the same matrix throughout)
        jumplimit: for how many timesteps after last being used are the state weights allowed to change
    """

    def __init__(self, n_regressors, T, prior_mean, P_0, Q, jumplimit=1):

        assert jumplimit > 0

        self.n_regressors = n_regressors
        self.T = T
        self.jumplimit = jumplimit
        self.x_0 = prior_mean
        self.P_0, self.Q = P_0, Q  # initial variance of regressors P_0, and variance for timesteps Q (can have different variances, we have it fixed)
        self.psi_diff_saves = []  # this can be used to resample the variance, but is currently unused
        self.noise_mean = np.zeros(self.n_regressors)  # save this, so as to not keep creating it
        self.identity = np.eye(self.n_regressors)  # not really needed, but kinda useful for state sampling

        self.weights = np.empty((self.T, self.n_regressors))
        # randomly initialise weights
        self.weights[0] = np.random.multivariate_normal(mean=self.x_0, cov=self.P_0)
        for t in range(1, T):
            self.weights[t] = self.weights[t - 1] + np.random.multivariate_normal(mean=self.noise_mean, cov=self.Q[t - 1])

    def rvs(self, inputs, times):
        """Given the input features and their time points, create responses from the dynamic GLM weights."""
        outputs = []
        for input, t in zip(inputs, times):
            if input.shape[0] == 0:
                output = np.zeros((0, 1))
            else:
                # find the distinct sets of features, how often they exist, and how to put the answers back in place
                types, inverses, counts = np.unique(input, return_inverse=True, return_counts=True, axis=0)

                # draw responses
                output = np.append(input, np.empty((input.shape[0], 1)), axis=1)
                for i, (type, c) in enumerate(zip(types, counts)):
                    temp = np.random.rand(c) < 1 / (1 + np.exp(- np.sum(self.weights[t] * type)))
                    output[inverses == i, -1] = temp
            outputs.append(output)
        return outputs

    def log_likelihood(self, input, timepoint):
        """
        Given input from a single session (a matrix containing features and responses) and the timepoint information from that session, return log-likelihoods for observations.
        """
        predictors, responses = input[:, :-1], input[:, -1]
        nans = np.isnan(responses)  # can come from cross-validation
        probs = np.zeros((input.shape[0], 2))
        out = np.zeros(input.shape[0])
        # I could possibly save the 1 / ..., since it's logged it's just - log (but the other half of the probs is an issue)
        probs[:, 1] = 1 / (1 + np.exp(- np.sum(self.weights[timepoint] * predictors, axis=1)))
        probs[:, 0] = 1 - probs[:, 1]
        # probably not necessary, just fill everything with probs and then have some be 1 - out?
        out[~nans] = probs[np.arange(input.shape[0])[~nans], responses[~nans].astype(int)]
        out = np.clip(out, np.spacing(1), 1 - np.spacing(1))
        out[nans] = 1

        return np.log(out)

    # Gibbs sampling
    def resample(self, data=[], testing_flag=False):
        """
        Resampling of dynamic logistic random variables. We get all the data that is assigned to a single state, and resample its random variables.
        We follow the resampling scheme of Windle: "Efficient Data Augmentation in Dynamic Models for Binary and Count Data".
        This makes use of the forward filter backwards sample algorithm. Which uses Kalman filtering, for which we use Anderson & Moore 1979

        Usually you just get one type of observation per timestep (one set of features), but we have a number of different feature sets per timestep. Thus,
        we invent more timesteps, but we don't allow the regressors to change from these steps to steps, effectively making the observations occur at the same
        time.
        """
        # TODO: Could clean up all this saving of info, so as to not have to constantly call model_save.delete_obs_data() in fit_script
        self.psi_diff_saves = []
        summary_statistics, all_times = self._get_statistics(data)
        types, pseudo_counts, counts = summary_statistics  # pseudo_counts are kappa_t in Windle scheme

        # if state is never active, resample from prior, but without dynamic change
        if len(counts) == 0:
            self.weights = np.tile(np.random.multivariate_normal(mean=self.x_0, cov=self.P_0), (self.T, 1))
            return

        # Sort out all the necessary timepoint variables (test case in dynamic_glm_dist_test.py)
        timepoint_map = {}  # this will ultimately map pseudo-weight onto real maps. Maps the ends of session (when all types (feature sets) have been processed) onto the correct real time points
        total_types = 0  # total count of relevant time steps: actual timesteps from different features in sessions + timesteps for change within jumplimit
        prev_t = all_times[0] - 1  # keep track of whether a jump happened (skipped sessions for this state)
        fake_times = []  # keep track of when we introduced fake times to accomodate changes on skipped sessions
        for type, t in zip(types, all_times):
            if t > prev_t + 1:  # if a jumnp happened, add fake timepoints, according to jumplimit
                fake_times += list(range(total_types, min(total_types + t - prev_t - 1, total_types + self.jumplimit)))
                for i, sub_t in enumerate(range(total_types, total_types + t - prev_t - 1)):
                    timepoint_map[prev_t + i + 1] = min(sub_t, total_types + self.jumplimit - 1)
                total_types += min(t - prev_t - 1, self.jumplimit)
            total_types += type.shape[0]
            timepoint_map[t] = total_types - 1
            prev_t = t

        # how many total different feature sets at different timepoints were observed
        actual_obs_count = sum([t.shape[0] for t in types])

        # We introduce pseudo-variance steps: we don't want change for the fake times, so only specific pseudo-Q's will be non-zero
        self.pseudo_Q = np.zeros((total_types, self.n_regressors, self.n_regressors))
        # is it okay to cut off last timepoint here?
        for k in range(self.T):
            if k in timepoint_map:
                self.pseudo_Q[timepoint_map[k]] = self.Q[k]  # for every real timepoint, map it's variance onto the pseudo_Q

        """Prepare for pseudo-observation sampling, collect all the psi's"""
        psis = np.empty(actual_obs_count)
        psi_counter = 0
        predictors = []
        for type, time in zip(types, all_times):
            for t in type:
                # psi (for a specific observation type t) is weights at time multiplied with the predictors t, i.e. the log odds
                psis[psi_counter] = np.sum(self.weights[time] * t)
                predictors.append(t)
                psi_counter += 1

        # Here we use the pg package to draw augmenting variable from a polya-gamma distribution
        # Windle calls the counts b_t = a_t + d_t, just the total number of times a psi has been seen. omega is the array in which our draws will be saved
        omega = np.empty(actual_obs_count)
        ppgs.pgdrawv(np.concatenate(counts).astype(float), psis, omega)

        # R is the variance on the observations in Kalman filtering. According to Windle, that is 1 / omega here
        self.R = np.zeros(total_types)
        mask = np.ones(total_types, dtype=bool)
        mask[fake_times] = False  # during fake times, we make no observations, so no variance
        self.R[mask] = 1 / omega

        # Create pseudo-obs. They can be treated as if they came from N(psi, 1 / omega)
        self.pseudo_obs = np.zeros(total_types)
        self.pseudo_obs[mask] = np.concatenate(pseudo_counts) / omega  # Windle: pseudo-data points are z_t = kappa_t / omega_t
        self.pseudo_obs = self.pseudo_obs.reshape(total_types, 1)

        # H is the matrix that transforms the latents into the (means of the) observations
        # in our case, these are the features, they multiply with the regressors to give us the means that we noisily observe
        self.H = np.zeros((total_types, self.n_regressors, 1))
        self.H[mask] = np.array(predictors).reshape(actual_obs_count, self.n_regressors, 1)

        """compute means and sigmas by filtering"""
        # if there is no obs, sigma_k = sigma_k_k_minus and x_hat_k = x_hat_k_k_minus (because R is infinite at that time)
        self.compute_sigmas(total_types)
        self.compute_means(total_types)

        """Sample states
        We follow Apendix 1 from "On Gibbs Sampling for State Space Models" by Carter & Kohn"""
        self.weights.fill(0)
        pseudo_weights = np.empty((total_types, self.n_regressors))
        # the first weight can be sampled directly, from the end of the Kalman inference process
        pseudo_weights[total_types - 1] = np.random.multivariate_normal(self.x_hat_k[total_types - 1], self.sigma_k[total_types - 1])

        # the other weights now depend on the weight sampled just after them, we need to update the Kalman estimates
        normals = np.random.standard_normal((total_types - 1, self.n_regressors))  # pregenerate some normals rvs to use later
        for k in range(total_types - 2, -1, -1):  # normally -1, but we already did first sampling
            if np.all(self.pseudo_Q[k] == 0):  # if this is a fake time, no change is allowed, just copy the weight
                pseudo_weights[k] = pseudo_weights[k + 1]
            else:
                # start with the Kalman filter inferences
                updated_x = self.x_hat_k[k].copy()  # not sure whether copy is necessary here
                updated_sigma = self.sigma_k[k].copy()

                # then update them with the subsequent weights
                for m in range(self.n_regressors):
                    # these are the Carter & Kohn equations, making use of the fact that F is identity, therefore x^tilde is x
                    # and delta_i is the variance Q
                    epsilon = pseudo_weights[k + 1, m] - updated_x[m]
                    state_R = updated_sigma[m, m] + self.pseudo_Q[k, m, m]

                    updated_x += updated_sigma[:, m] * epsilon / state_R
                    updated_sigma -= updated_sigma.dot(np.outer(self.identity[m], self.identity[m])).dot(updated_sigma) / state_R

                pseudo_weights[k] = local_multivariate_normal_draw(updated_x, updated_sigma, normals[k])  # transform our normal according to pre-computed mean and var

        # put the pseudo_weights where they belong
        for k in range(self.T):
            if k in timepoint_map:
                self.weights[k] = pseudo_weights[timepoint_map[k]]

        # Sample before and after active times too
        # fill weights up before first determined weight, at least until the jumplimit is reached (in principle this would need to be shrunk by P_0)
        for k in range(all_times[0] - 1, -1, -1):
            if k > all_times[0] - self.jumplimit - 1:
                self.weights[k] = self.weights[k + 1] + np.random.multivariate_normal(self.noise_mean, self.Q[k])
            else:
                self.weights[k] = self.weights[k + 1]
        # fill weights up after last determined weights, at least until jumplimit
        for k in range(all_times[-1] + 1, self.T):
            if k < min(all_times[-1] + 1 + self.jumplimit, self.T):
                self.weights[k] = self.weights[k - 1] + np.random.multivariate_normal(self.noise_mean, self.Q[k])
            else:
                self.weights[k] = self.weights[k - 1]

        if testing_flag:
            return timepoint_map
        return pseudo_weights
        # If one wants to resample variance...
        # self.psi_diff_saves = np.concatenate(self.psi_diff_saves)

    def _get_statistics(self, data):
        """For the data of a list of sessions, we need to know:
        which unique features did the state encounter (types)
        what is their corresponding pseudo-observation (pseudo_counts)
        how often did they occur (counts)
        also return which timepoints had non-zero data (times)
        """
        summary_statistics = [[], [], []]
        times = []
        if isinstance(data, np.ndarray):
            warn('What you are trying is probably stupid, at least the code is not implemented')
            quit()
        else:
            for i, d in enumerate(data):
                clean_d = d[~np.isnan(d[:, -1])]
                if len(clean_d) != 0:
                    predictors, responses = clean_d[:, :-1], clean_d[:, -1]
                    types, inverses, counts = np.unique(predictors, return_inverse=True, return_counts=True, axis=0)
                    pseudo_counts = np.zeros(len(types))
                    for j, c in enumerate(counts):
                        mask = inverses == j
                        # find out how the animals answered on these trials, and turn it into a pseudo-observation
                        pseudo_counts[j] = np.sum(responses[mask]) - c / 2  # already turn this into the desired format: kappa_t = a_t -b_t / 2
                    summary_statistics[0].append(types)
                    summary_statistics[1].append(pseudo_counts)
                    summary_statistics[2].append(counts)
                    times.append(i)

        return summary_statistics, times

    def compute_sigmas(self, T):
        """Sigmas can be precomputed (without z), we do this here."""
        # We rely on the fact that H.T.dot(sigma).dot(H) is just a number, no matrix inversion needed
        # furthermore we use the fact that many matrices are identities, namely F and G
        # We follow Anderson & Moore 1979, p.44
        self.sigma_k = []  # we have to reset this for repeating this calculation later for the resampling (R changes)
        self.sigma_k_k_minus = [self.P_0]
        self.gain_save = []
        for k in range(T):
            if self.R[k] == 0:  # these are fake times, no observations were made but we allow change as per the jumplimit
                self.gain_save.append(None)  # need a placeholder
                self.sigma_k.append(self.sigma_k_k_minus[k])
                self.sigma_k_k_minus.append(self.sigma_k[k] + self.pseudo_Q[k])  # the amount of allowed change just adds up if no obs
            else:
                sigma, H = self.sigma_k_k_minus[k], self.H[k]  # we will need this a lot, so shorten it
                gain = sigma.dot(H).dot(1 / (H.T.dot(sigma).dot(H) + self.R[k]))  # matrix K, elsewhere called gain
                self.gain_save.append(gain)  # the gain matrix is useful for compute_means as well, save it

                # the two main equations, from p.44:
                self.sigma_k.append(sigma - gain.dot(H.T).dot(sigma))
                self.sigma_k_k_minus.append(self.sigma_k[k] + self.pseudo_Q[k])

    def compute_means(self, T):
        """Compute the means, the estimates of the states.
        Used to also contain self.x_hat_k_k_minus, but it's not necessary for our setup
        (Especially (if not generally), with F=1, x_hat_k_k_minus is just equal to x_hat_k, see also p.40)"""
        # We follow Anderson & Moore 1979, p.44
        self.x_hat_k = [self.x_0]  # we have to reset this for repeating this calculation later for the resampling
        for k in range(T):  # this will leave out last state which doesn't have observation
            if self.gain_save[k] is None:
                self.x_hat_k.append(self.x_hat_k[k])  # mean stays unchanged if no observations
            else:
                x, H = self.x_hat_k[k], self.H[k]  # we will need this a lot, so shorten it
                self.x_hat_k.append(x + self.gain_save[k].dot(self.pseudo_obs[k] - H.T.dot(x)))

        self.x_hat_k.pop(0)  # remove initialisation element from list

    def num_parameters(self):
        # Compute WAIC in principle
        return self.weights.size

    # Max likelihood
    def max_likelihood(self, data, weights=None):
        warn('ML not implemented')

    def MAP(self, data, weights=None):
        warn('MAP not implemented')
